' ____________________________________________________________________________
'
' TidyMP3 v1.4 - By BaSs-Z (C) 2004 - Public Domain! www.cuemaster.host.sk
'
' Sourcecode for PowerBASIC 32-bit Console Compiler 2.10 for Windows
'
' You are free to do whatever you want with this sourcecode. However, if you
' decide to use it, I would appreciate it if you mention my name in your
' aboutbox or readmefile.
' ____________________________________________________________________________


$Compile Con "TIDYMP3.EXE"
$Dim All
$Debug Error Off

%MP3MAXHEADER = 64000   ' maximum offset of mp3 header (excluding ID3v2-tag)
%FRAMES_FLAG = &h0001   ' xing-header frames-flag

%TRUE = 1
%FALSE = 0
%SUBDIR = 16

' Win32Api declarations
Type FILETIME
  dwLowDateTime As Dword
  dwHighDateTime As Dword
End Type
Declare Function GetFileTime Lib "KERNEL32.DLL" Alias "GetFileTime" _
  (ByVal hFile As Dword, lpCreationTime As FILETIME, _
   lpLastAccessTime As FILETIME, lpLastWriteTime As FILETIME) As Long
Declare Function SetFileTime Lib "KERNEL32.DLL" Alias "SetFileTime" _
  (ByVal hFile As Dword, lpCreationTime As FILETIME, _
   lpLastAccessTime As FILETIME, lpLastWriteTime As FILETIME) As Long
Declare Function SetConsoleCtrlHandler Lib "KERNEL32.DLL" Alias _
  "SetConsoleCtrlHandler" (ByVal HandlerRoutine As Dword, ByVal _
  nAdd As Long) As Long
%ENABLE_PROCESSED_INPUT = &H1
Declare Function GetConsoleMode Lib "KERNEL32.DLL" Alias "GetConsoleMode" _
  (ByVal hConsoleHandle As Dword, lpMode As Dword) As Long
Declare Function SetConsoleMode Lib "KERNEL32.DLL" Alias "SetConsoleMode" _
  (ByVal hConsoleHandle As Dword, ByVal dwMode As Dword) As Long
%CTRL_C_EVENT        = 0
%CTRL_BREAK_EVENT    = 1
%CTRL_CLOSE_EVENT    = 2
%CTRL_LOGOFF_EVENT   = 5
%CTRL_SHUTDOWN_EVENT = 6

Global BreakFlag As Long

' Command-line:
Global argc As Long
Global argv() As String

' Command-line options:
Global optVerbose As Long
Global optSaveBackup As Long
Global optSaveTrash As Long
Global optDisFrameClean As Long
Global optCompatMode As Long
Global optSubDirs As Long
Global optPreserveID3v1 As Long
Global optPreserveTimeStamp As Long

' Statistics:
Global FileCount As Long
Global TotalCleaned As Quad

' MP3 header info arrays:
Global VersionArr() As Long
Global LayerArr() As Long
Global BitrateMatrix() As Long
Global FreqArr() As Long

' This will contain the MP3-header info for the current file:
Type TMP3Info
  FileSize     As Quad  ' Total filesize of the MP3-file
  Frames       As Dword ' Total number of frames in the MP3-file
  ID3v2Tag     As Long  ' Has ID3v2-tag %TRUE / %FALSE
  ID3v2TagSize As Dword ' Size of the ID3v2-tag
  VBR          As Long  ' Variable Bitrate file %TRUE / %FALSE
  head(0 To 3) As Byte  ' Current MPEG-header (32 bits)
  Frequency    As Long  ' Sampling frequency
  Bitrate      As Long  ' Bitrate (kbps)
  Padding      As Long  ' Padding bit %TRUE / %FALSE
  MPEGVersion  As Long  ' MPEG-version ID
  MPEGLayer    As Long  ' MPEG-layer
  FrameSize    As Long  ' Size of the current frame (bytes)
End Type
Global MP3Info As TMP3Info

' Sub / Function declarations:
Declare Sub FillMP3Arrays()
Declare Sub ParseCmdLine()
Declare Sub DoFiles(sFileSpec As String)
Declare Sub DoFile(sFileName As String)
Declare Function GetID3v1Tag(fn As Long) As String
Declare Function CheckRIFF(fn As Long) As Long
Declare Sub SkipID3v2Tag(FileNum As Long)
Declare Sub CheckFirstFrame()
Declare Function CleanFrame(ByRef sFrameData As String) As Long
Declare Function TrimFile(sPathName As String) As String
Declare Function TrimFile(sPathName As String) As String
Declare Function MPEGVersionStr(vID As Long) As String
Declare Function ExtractI4(lpBuf As Byte Ptr) As Long
Declare Function FindFrameHeader(FileNum As Long, FirstFrame As Long) As Long
Declare Function DecodeFrameHeader() As Long
Declare Function IsID3v1Tag(ByRef sTagData As String) As Long
Declare Function IsID3v1TagField(ByRef sTagData As String) As Long


' Main function:
Function PBMain
  Dim bShowHelp As Long
  Dim sTmp As String
  Dim i As Long, j As Long
  Dim res As Long
  Dim sFile As String
  Dim sFiles As String
  Dim dwConsMode As Dword


  StdOut "TidyMP3 v1.4(mod/1) - By BaSs-Z (C) 2004 - Public Domain! www.cuemaster.host.sk"
  StdOut "Disclaimer: Author cannot be held responsible for any damage."
  StdOut "Always make backups when working with important data!
  StdOut

  ParseCmdLine

  ' Handle command line:
  If argc = 1 Then
    bShowHelp = %TRUE
  Else
    For i = 1 To argc - 1
      res = %FALSE
      If Left$(argv(i), 2) = "--" Then
        sTmp = Mid$(argv(i), 3)
        res = %TRUE
      ElseIf Left$(argv(i), 1) = "/" Or Left$(argv(i), 1) = "-" Then
        sTmp = Mid$(argv(i), 2)
        res = %TRUE
      Else
        res = %FALSE
      End If

      If res Then
        For j = 1 To Len(sTmp)
          Select Case LCase$(Mid$(sTmp, j, 1))
            Case "?", "h": bShowHelp = %TRUE
            Case "v": optVerbose = %TRUE
            Case "b": optSaveBackup = %TRUE
            Case "t": optSaveTrash = %TRUE
            Case "f": optDisFrameClean = %TRUE
            Case "c": optCompatMode = %TRUE
            Case "s": optSubDirs = %TRUE
            Case "p": optPreserveID3v1 = %TRUE
            Case "d": optPreserveTimeStamp = %TRUE
            Case Else
              StdOut "Invalid switch: " & argv(i)
              Exit Function
          End Select
        Next j
      Else
        ' Add file(spec) to list:
        If Len(sFiles) Then
          sFiles = sFiles & "|"
        End If
        sFiles = sFiles & argv(i)
      End If
    Next i
  End If

  ' Show syntax if neccessary:
  If bShowHelp Then
    StdOut "Removes all trash (ID3-tags, images, comments, lyrics or broken frames) from"
    StdOut "mp3 files. Advanced mp3-frame scanner which removes all non-frame data. This"
    StdOut "also fixes any buffer-overflow bugs with WindowsXP or WinAMP."
    StdOut
    StdOut "Syntax: tidymp3 filespec1 [filespec2 [...]] {options}"
    StdOut
    StdOut "Options:"
    StdOut "  /V   Verbose mode                    /F   Disable frame-reconstructor"
    StdOut "  /B   Save backup to filename.bak     /C   Compatibility mode (use tempfile)"
    StdOut "  /T   Save trash to filename.trash    /S   Recurse subfolders"
    StdOut "  /P   Preserve ID3v1 tags only        /D   Preserve modification timestamp"
    StdOut "See README.TXT for detailed information about these options."
    StdOut
    StdOut "Notes:"
    StdOut "1) If the file is not an mp3 at all, it's left untouched."
    StdOut "2) 100% compatible with variable bitrate files."
    StdOut "3) An mp3-encoded WAV-file is converted to MP3-file (RIFF/WAVE-header removed)."
    StdOut
    StdOut "Example: tidymp3 ""d:\my mp3z\*.mp3"" c:\download\mysong.mp3 /T"
    Exit Function
  End If

  If optSaveBackup Then optCompatMode = %TRUE

  ' Display active options:
  If optVerbose Then              StdOut "* Verbose mode"
  If optSaveBackup Then           StdOut "* Save backups"
  If optSaveTrash Then            StdOut "* Save trash-files"
  If optDisFrameClean Then        StdOut "* Frame-reconstructor disabled"
  If optCompatMode Then           StdOut "* Compatibility mode"
  If optSubDirs Then              StdOut "* Recurse subfolders"
  If optPreserveID3v1 Then        StdOut "* Preserve ID3v1 tags only"
  If optPreserveTimeStamp Then    StdOut "* Preserve modification timestamps"

  StdOut "Use /? option for help."
  StdOut

  ' Prepare the MP3-header info arrays:
  Call FillMP3Arrays

  FileCount = 0
  TotalCleaned = 0

  ' Activate Ctrl+C and Ctrl+Break handler:
  If GetConsoleMode(GetStdIn, dwConsMode) Then
    SetConsoleMode GetStdIn, (dwConsMode Or %ENABLE_PROCESSED_INPUT)
  End If
  SetConsoleCtrlHandler CodePtr(CtrlHandler), %TRUE

  ' Parse filelist:
  While Len(sFiles)
    sFile = Extract$(sFiles, "|")
    If InStr(sFiles, "|") Then
      sFiles = Remain$(sFiles, "|")
    Else
      sFiles = ""
    End If

    If BreakFlag Then Exit Loop
    DoFiles sFile
  Wend

  ' Deactivate Ctrl+C and Ctrl+Break handler:
  SetConsoleCtrlHandler CodePtr(CtrlHandler), %FALSE

  StdOut
  StdOut "Finished! Processed " & Format$(FileCount) & " file(s) - " & _
         "cleaned " & Format$(TotalCleaned) & " bytes"
End Function

' Parse the COMMAND$ variable into argc and argv()
Sub ParseCmdLine()
  Local i As Long, c As String, q As Long, cmd As String
  ReDim argv(0 To 1)
  cmd = Command$
  argc = 1
  q = %FALSE
  For i = 1 To Len(cmd)
    c = Mid$(cmd, i, 1)
    Select Case c
      Case " "
        If IsFalse q Then
          If Len(argv(argc)) Then
            Incr argc
            ReDim Preserve argv(0 To argc)
          End If
        Else
          argv(argc) = argv(argc) & c
        End If
      Case """"
        q = IsFalse q
      Case Else
        argv(argc) = argv(argc) & c
    End Select
  Next i
  If Len(argv(argc)) Then Incr argc
End Sub


' Removes the filename part from a pathname.
' "C:\TEST\FILE.TXT" will return "C:\TEST\"
' "FILE.TXT" will return an empty string
Function TrimFile(sPathName As String) As String
  Local i As Long
  For i = Len(sPathName) To 1 Step -1
    If Mid$(sPathName, i, 1) = "\" Then
      Function = Left$(sPathName, i)
      Exit For
    End If
  Next i
End Function


' Removes the pathname part from a filename.
' "C:\TEST\FILE.TXT" will return "FILE.TXT"
' "FILE.TXT" will return "FILE.TXT"
Function TrimPath(s As String) As String
  Local i As Long
  For i = Len(s) To 1 Step -1
    If Mid$(s,i,1)="\" Then
      Function = Mid$(s, i+1)
      Exit Function
    End If
  Next i
  Function = s
End Function


' Parse filespec and process found files
Sub DoFiles(sFileSpec As String)
  Dim s As String, attr As Long, i As Long
  Dim sFileList() As String, lFileCount As Long

  s = Dir$(sFileSpec)
  While Len(s)
    s = TrimFile(sFileSpec) & s
    attr = GetAttr(s) And %SUBDIR
    If (attr = 0) Then
      Incr lFileCount
      ReDim Preserve sFileList(0 To lFileCount)
      sFileList(lFileCount) = s
    End If
    s = Dir$
  Wend

  If optSubDirs Then
    s = Dir$(TrimFile(sFileSpec) & "*.*", %SUBDIR)
    While Len(s)
      s = TrimFile(sFileSpec) & s
      attr = GetAttr(s) And %SUBDIR
      If attr Then
        Incr lFileCount
        ReDim Preserve sFileList(0 To lFileCount)
        sFileList(lFileCount) = s & "\"
      End If
      s = Dir$
    Wend
  End If

  For i = 1 To lFileCount
    If Right$(sFileList(i), 1) = "\" Then
      DoFiles sFileList(i) & TrimPath(sFileSpec)
    Else
      DoFile sFileList(i)
    End If
    If BreakFlag Then Exit For
  Next i
End Sub


' Returns the MPEG-version as a user-readable string.
Function MPEGVersionStr(vID As Long) As String
  Select Case vID
    Case 1 : Function = "1.0"
    Case 2 : Function = "2.0"
    Case 25: Function = "2.5"
    Case Else: Function = Format$(vID)
  End Select
End Function


Sub MyPut(ByVal fn As Long, ByRef buf As String)
  Static mybuf As String, writepos As Dword
  Local dif As Long, s As String, readpos As Dword
  If fn = 0 Then  ' reset
    mybuf = ""
    writepos = 0
  Else
    mybuf = mybuf & buf
    If Len(mybuf) Then
      readpos = Seek(fn) - 1
      dif = readpos - writepos
      If dif>0 Then
        Seek #fn, writepos + 1
        s = Left$(mybuf, dif)
        dif = Len(mybuf) - Len(s)
        If dif >= 0 Then mybuf = Right$(mybuf, dif)
        Put$ #fn, s
        writepos = writepos + Len(s)
        Seek #fn, readpos + 1
      End If
    End If
  End If
End Sub

Function CtrlHandler(ByVal dwCtrlType As Dword) As Long
  Dim sEvent As String

  Select Case dwCtrlType
    Case %CTRL_C_EVENT:        sEvent = "CTRL_C_EVENT"
    Case %CTRL_BREAK_EVENT:    sEvent = "CTRL_BREAK_EVENT"
    Case %CTRL_CLOSE_EVENT:    sEvent = "CTRL_CLOSE_EVENT"
    Case %CTRL_LOGOFF_EVENT:   sEvent = "CTRL_LOGOFF_EVENT"
    Case %CTRL_SHUTDOWN_EVENT: sEvent = "CTRL_SHUTDOWN_EVENT"
    Case Else: Exit Function
  End Select

  If optCompatMode Then
    StdErr
    StdErr "*** " & sEvent & " -> Canceling current file ***"
  Else
    StdErr
    StdErr "*** " & sEvent & " -> Finishing current file ***"
  End If
  BreakFlag = %TRUE
  Function = %TRUE
End Function


' Process one file:
Sub DoFile(sFileName As String)
  Local res As Long
  Local frame As Long
  Local fn As Long, fnOut As Long, fnOutTrash As Long
  Local sTmp As String
  Local FrameSize As Dword
  Local FrameData As String
  Local OutWritten As Quad
  Local inPos As Quad
  Local tSize As Quad
  Local FramesCount As Dword
  Local hFile As Dword, ID3v1Tag As String
  Dim FileTimes(0 To 2) As FILETIME

  StdOut sFileName;
  res = %FALSE

  ' frame fixed flag
  frame = 0

  ' reset delayed file-write function:
  MyPut 0, ""

  fn = FreeFile
  If optCompatMode Then
    ' compatibility mode: only READ from inputfile, write to temporary file
    Open sFileName For Binary Access Read Lock Shared As #fn
  Else
    ' normal mode: READ and WRITE to the same file
    Open sFileName For Binary Access Read Write Lock Write As #fn
  End If
  If ErrClear Then
    StdOut "...Cannot open!"
  Else
    MP3Info.FileSize = Lof(fn)

    If optPreserveTimeStamp Then
      hFile = FileAttr(fn, 2)
      If GetFileTime(hFile, FileTimes(0), FileTimes(1), FileTimes(2)) = 0 Then
        ' failed...
        hFile = 0
      End If
    End If

    If CheckRIFF(fn) = -1 Then
      Close #fn
      Exit Sub
    End If

    If optPreserveID3v1 Then
      ID3v1tag = GetID3v1Tag(fn)
    End If

    SkipID3v2Tag fn
    If IsFalse FindFrameHeader(fn, %TRUE) Then
      StdOut "...No MP3!"
    Else
      If optVerbose Then
        StdOut
        StdOut " - MPEG " & MPEGVersionStr(MP3Info.MPEGVersion) & _
                 " layer " & Format$(MP3Info.MPEGLayer)
        StdOut " - Header found at: " & Format$(Seek(fn) - 1) & " bytes"
        If MP3Info.VBR Then
          StdOut " - XingVBR - Total number of frames = " & _
                               Format$(MP3Info.Frames)
        End If
      End If

      If optCompatMode Then
        fnOut = FreeFile
        Open sFileName & ".clean" For Binary Access Write Lock Write As #fnOut
        If Lof(fnOut) Then SetEof #fnOut  ' empty file if neccessary
      End If

      If optSaveTrash Then
        fnOutTrash = FreeFile
        Open sFileName & ".trash" For Binary Access Write _
                                             Lock Write As #fnOutTrash
        If Lof(fnOutTrash) Then SetEof #fnOutTrash ' empty file if neccessary
      End If
      FramesCount = 0
      OutWritten = 0
      inPos = 1
      Do
        If optCompatMode And BreakFlag Then Exit Sub

        Incr FramesCount

        ' Copy trash if neccessary:
        If optSaveTrash Then
          If Seek(fn) > inPos Then
            tSize = Seek(fn) - inPos
            Seek #fn, inPos
            Get$ #fn, tSize, sTmp
            Put$ #fnOutTrash, sTmp
          End If
        End If

        ' Show status if neccessary:
        If optVerbose Then
          StdOut $Cr & _
                 " - Frame #" & Format$(FramesCount, "* ###") & ":" & _
                 Format$(MP3Info.FrameSize, "* ##") & " bytes " & _
                 Format$(MP3Info.Bitrate, "* #") & "kbit " & _
                 Format$(MP3Info.Frequency, "* ###") & "Hz";
        End If

        ' Read in the MP3-frame:
        Get$ #fn, MP3Info.FrameSize, FrameData
        If Len(FrameData) = MP3Info.FrameSize Then
          'clean it:
          frame = frame + CleanFrame(FrameData)
        
          If Len(FrameData) = MP3Info.FrameSize Then
            If optCompatMode Then
              'write to outfile:
              Put$ #fnOut, FrameData
            Else
              MyPut fn, FrameData
            End If
          End If
          
          ' Update variables:
          OutWritten = OutWritten + MP3Info.FrameSize
        
        Else
          ' incomplete framedata (probably EOF)
          If optSaveTrash Then
            Put$ #fnOutTrash, FrameData
          End If
        End If
        inPos = Seek(fn)

        ' FrameCount-limit if VBR
        If IsTrue MP3Info.VBR Then
          If FramesCount >= MP3Info.Frames Then
            Exit Do
          End If
        End If

      ' Find next frameheader:
      Loop While FindFrameHeader(fn, %FALSE)

      ' Copy leftover trash if neccessary:
      If optSaveTrash Then
        Get$ #fn, Lof(fn) - Seek(fn) + 1, sTmp
        Put$ #fnOutTrash, sTmp
        If Lof(fnOutTrash) = 0 Then
          ' Trashfile is empty, so delete it...
          sTmp = FileName$(fnOutTrash)
          Close #fnOutTrash
          Kill sTmp
        Else
          Close #fnOutTrash
        End If
      End If

      If MP3Info.VBR And (MP3Info.Frames > FramesCount) Then
        ' TODO: Reconstruct VBR-header, because it contains too
        '       low framecount. (Additional frames found)
      End If

      If optCompatMode Then
        sTmp = FileName$(fnOut)

        If optPreserveID3v1 And (Len(ID3v1tag) = 128) Then
          Put$ #fnOut, ID3v1tag
          OutWritten = OutWritten + 128
        End If

        ' update file time:
        If optPreserveTimeStamp And (hFile <> 0) Then
          hFile = FileAttr(fnOut, 2)
          SetFileTime hFile, FileTimes(0), FileTimes(1), FileTimes(2)
        End If

        ' Close output file
        Close #fnOut

      Else
        ' write id3v1 tag
        If optPreserveID3v1 And (Len(ID3v1tag) = 128) Then
          MyPut fn, ID3v1tag
          OutWritten = OutWritten + 128
        End If

        Seek #fn, OutWritten + 1
        ' write pending data
        MyPut fn, ""
        ' set end-of-file marker:
        SetEof #fn
        ' update file time:
        If optPreserveTimeStamp And (hFile <> 0) Then
          SetFileTime hFile, FileTimes(0), FileTimes(1), FileTimes(2)
        End If
      End If

      ' Show statistics if neccessary:
      If optVerbose Then
        StdOut
        StdOut " - " & Format$(MP3Info.FileSize) & " bytes read - " & _
                       Format$(OutWritten) & " bytes saved"
        If IsFalse optDisFrameClean Then
           StdOut " - " & Format$(frame) & " frame(s) fixed"
        End If
      End If

      res = %TRUE
    End If
    Close #fn

  End If

  If res Then
    If optCompatMode Then
      If optSaveBackup Then
        Kill sFileName & ".bak"
        ErrClear
        ' rename original to .BAK
        Name sFileName As sFileName & ".bak"
      Else
        ' delete original
        Kill sFileName
      End If
      ' rename tempfile to original filename
      Name sTmp As sFileName
      If Err Then StdOut " ! Could not rename " & sTmp & " to " & sFileName

    End If 'optCompatMode

    StdOut " - Cleaned " & Format$(MP3Info.FileSize - OutWritten) & " bytes"
    TotalCleaned = TotalCleaned + MP3Info.FileSize - OutWritten
    Incr FileCount
  End If
End Sub

Function GetID3v1Tag(fn As Long) As String
  Dim fp As Dword, fs As Dword, s As String
  fs = Lof(fn)
  If fs >= 128 Then
    fp = Seek(fn)
    Seek #fn, fs - 127
    Get$ #fn, 128, s
    Seek #fn, fp
    If IsFalse IsID3v1Tag(s) Then s = ""
  End If
  Function = s
End Function


' Check for RIFF-wave-header (.WAV-file):
' returns -1 if error
'         0 if no RIFF
'         1 if RIFF header skipped
Function CheckRIFF(fn As Long) As Long
  Dim s As String, s2 As String, d As Dword, p As Dword, fs As Dword, rc As Long
  If Seek(fn) <> 1 Then Seek #fn, 1
  Get$ #fn, 4, s
  If s = "RIFF" Then
    fs = Lof(fn)
    Get$ #fn, 8, s  ' riffsize[dword] & "WAVE"

    rc = -1
    If Right$(s,4) <> "WAVE" Then
      StdOut "...RIFF but not WAVE (" & Right$(s,4) & ")!"
    Else

      p = 13
      Do
        Get$ #fn, 4, s
        Get #fn, , d
        p=p+8+d
        If s = "data" Then
          rc = 1
          Exit Do
        End If
        Seek #fn, p
      Loop Until p>=fs

      If rc = -1 Then
        StdOut "...WAVE-file chunk '" & s & "' incomplete, 'data' not found!"
      End If

    End If

  Else
    ' move back filepointer:
    Seek #fn, 1
    rc = 0
  End If
  Function = rc
End Function

' Looks for an ID3v2-tag and places the filepointer exactly after it.
Sub SkipID3v2Tag(fn As Long)
  Dim TagMagic As String, sp As Dword
  Dim TagSize As Dword, i As Long
  Dim TagHead As String, bp As Byte Ptr
  sp = Seek(fn)
  MP3Info.ID3v2Tag = %FALSE
  TagSize = 0
  Get$ #fn, 3, TagMagic
  If TagMagic = "ID3" Then
    Get$ #fn, 7, TagHead
    bp = StrPtr(TagHead)
    If (@bp[0] < &hFF) And (@bp[1] < &hFF) And _
       (@bp[3] < &h80) And (@bp[4] < &h80) And _
       (@bp[5] < &h80) And (@bp[6] < &h80) Then
      MP3Info.ID3v2Tag = %TRUE

      ' calculate ID3v2-tag-size:
      TagSize = @bp[3] * &h200000 + _
                @bp[4] * &h4000 + _
                @bp[5] * &h80 + _
                @bp[6] + 10                     ' 10=header
      ' check if footer-bit is set:
      If (@bp[2] And &h10) Then TagSize = TagSize + 10
    End If
  End If
  ' skip the tag:
  Seek #fn, sp + TagSize
  MP3Info.ID3v2TagSize = TagSize
End Sub


' big endian extract (for Xing header)
Function ExtractI4(buf As Byte Ptr) As Long
  Static x As Long
  x = @buf[0]
  Shift Left x, 8
  x = x Or @buf[1]
  Shift Left x, 8
  x = x Or @buf[2]
  Shift Left x, 8
  x = x Or @buf[3]
  Function = x
End Function


' Searches an open file for the next mp3 frame-header and places filepointer
' at header-start. Stops after %MP3MAXHEADER bytes or at EOF. If no header is
' found, %FALSE is returned.
Function FindFrameHeader(fn As Long, FirstFrame As Long) As Long
  Static i As Long, rd As Long         ' all static varbs for highspeed :)
  Static buf As String, s As String
  Static p As Quad
  Static res As Long
  Static b As Byte Ptr
  'for Xing-stuff:
  Static h_id As Byte, h_mode As Byte
  Static offs As Long, head_flags As Long

  buf = String$(8, 0)
  rd = 0
  res = %FALSE
  p = Seek(fn)       ' store filepointer
  Do
    ' read 4 bytes from file:
    Get$ #fn, 4, s
    ' add it to the 8-byte buffer:
    Mid$(buf, 1, 4) = Mid$(buf, 5, 4)
    Mid$(buf, 5, 4) = s

    ' check if the buffer contains a frame-header:
    i = InStr(buf, Chr$(255))             ' all bits set
    If i > 0 And i < 6 Then
      If (Asc(Mid$(buf, i+1, 1)) And &h70) = &h70 Then    ' last 3 bits set
        ' OK: valid frame-header!
        b = StrPtr(buf)
        MP3Info.head(0) = @b[i-1]
        MP3Info.head(1) = @b[i]
        MP3Info.head(2) = @b[i+1]
        MP3Info.head(3) = @b[i+2]

        res = DecodeFrameHeader()

        If res And FirstFrame Then
          ' check for Xing-VBR
          MP3Info.VBR = %FALSE

          h_id       = (@b[i] And &h18) \ 8    ' (buf[1] >> 3) & 1
          h_mode     = (@b[i+2] And &hC0) \ 64 ' (buf[3] >> 6) & 3;

          ' determine offset of header
          If h_id Then ' mpeg1
            If h_mode <> 3 Then offs = 32 + 4 _
            Else                offs = 17 + 4
          Else         ' mpeg2
            If h_mode <> 3 Then offs = 17 + 4 _
            Else                offs =  9 + 4
          End If

          Seek #fn, p + rd + (i - 5) + offs
          Get$ #fn, 4, s
          If s = "Xing" Then
            ' VBR-header! save number of frames, and stop copying after these
            ' number of frames this will prevent copying of trash-data hidden
            ' in valid frames after the mp3
            Get$ #fn, 4, s    ' get flags
            b = StrPtr(s)
            head_flags = ExtractI4(b)
            If (head_flags And %FRAMES_FLAG) Then
              Get$ #fn, 4, s  ' get number of frames
              MP3Info.Frames = ExtractI4(StrPtr(s))
              MP3Info.VBR = %TRUE
            End If
          End If

        End If

        ' set filepointer to beginning of header
        Seek #fn, p + rd + (i - 5)
      End If
    End If

    rd = rd + 4
  Loop Until Eof(fn) Or (rd > %MP3MAXHEADER) Or IsTrue res

  If IsFalse res Then
    ' nothing found so set filepointer all the way back:
    Seek #fn, p
  End If

  Function = res
End Function


' Fills the global arrays neccessary for DecodeFrameHeader()
Sub FillMP3Arrays()
  Local i As Long

  ReDim VersionArr(0 To 3)
  VersionArr(0) = 25:   VersionArr(2) = 2
  VersionArr(1) = -1:   VersionArr(3) = 1

  ReDim LayerArr(0 To 3)
  LayerArr(0) = -1:     LayerArr(2) = 2
  LayerArr(1) = 3 :     LayerArr(3) = 1

  ReDim FreqArr(0 To 3, 0 To 3)
  ' MPEG2.5:
  FreqArr(0, 0) = 11025: FreqArr(0, 1) = 12000: FreqArr(0, 2) = 8000
  ' MPEG2:
  FreqArr(2, 0) = 22050: FreqArr(2, 1) = 24000: FreqArr(2, 2) = 16000
  ' MPEG1:
  FreqArr(3, 0) = 44100: FreqArr(3, 1) = 48000: FreqArr(3, 2) = 32000
  ' values that are left out stay 0, which means Invalid

  ReDim BitrateMatrix(0 To 79)
  For i = 0 To 79
    BitrateMatrix(i) = Val(Read$(i + 1))
  Next i
  Data 0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,-1
  Data 0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,-1
  Data 0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,-1
  Data 0,32,48,56, 64, 80, 96,112,128,144,160,176,192,224,256,-1
  Data 0, 8,16,24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,-1
End Sub


' Decodes the frameheader in MP3Info.head() and stores the
' frameinfo in MP3Info.xxx
Function DecodeFrameHeader() As Long
  Static v As Byte, x As Byte, b As Byte

  ' MPEGVersion:
  v = (MP3Info.head(1) And &h18) \ 8    ' bit 3 and 4
  MP3Info.MPEGVersion = VersionArr(v)
  If MP3Info.MPEGVersion = -1 Then Exit Function

  ' MPEG Layer:
  x = (MP3Info.head(1) And &h6) \ 2     ' bits 1 and 2
  MP3Info.MPEGLayer = LayerArr(x)
  If MP3Info.MPEGLayer = -1 Then Exit Function

  ' Padding bit:
  x = (MP3Info.head(2) And &h2) \ 2     ' bit 1
  MP3Info.Padding = x

  ' Sampling rate frequency:
  x = (MP3Info.head(2) And &hC) \ 4     ' bits 2 and 3
  MP3Info.Frequency = FreqArr(v, x)
  If MP3Info.Frequency = 0 Then Exit Function

  ' Bitrate (for this frame!):
  If MP3Info.MPEGVersion = 1 Then
    b = MP3Info.MPEGLayer - 1
  Else 'MPEG2 or MPEG2.5
    If MP3Info.MPEGLayer=1 Then b = 3 Else b = 4
  End If
  b = b * 16
  x = (MP3Info.head(2) And &hF0) \ &h10 ' bits 3-7
  MP3Info.Bitrate = BitrateMatrix(x+b)
  If MP3Info.Bitrate < 1 Or MP3Info.Bitrate > 600 Then Exit Function

  ' Calculate framesize (bytes)
  MP3Info.FrameSize = _
    Int((CSng(MP3Info.Bitrate) * 144000) / CSng(MP3Info.Frequency))
  If MP3Info.Padding Then
    If MP3Info.MPEGLayer = 1 Then
      ' For Layer I padding is 32 bits (4 bytes) long
      MP3Info.FrameSize = MP3Info.FrameSize + 4
    Else
      ' for Layer II and Layer III padding is 8 bits (1 byte) long
      MP3Info.FrameSize = MP3Info.FrameSize + 1
    End If
  End If

  Function = %TRUE
End Function


' Cleans a MP3 frame: searches for leftover tag-/lyric-data and reconstructs
' the original audiodata
Function CleanFrame(ByRef FrameData As String) As Long
  Local res As Long
  res = 0

  If IsTrue optDisFrameClean Then Function = 0
  ' yes, they really exist: 'Tagging-tools' that actually OVERWRITE the
  ' frames' audiodata
  Static tmpStr As String, i As Long, fs As Long
  fs = Len(FrameData)
  i = InStr(FrameData, "TAG")
  If i Then
    tmpStr = Mid$(FrameData, i)
    If Len(tmpStr) < 128 Then
      res = 1
      tmpStr = tmpStr & String$(128 - Len(tmpStr), 0) 'pad with zeroes
    End If
    If IsID3v1Tag(tmpStr) Then
      ' hmmmz another stupid tag-util, try to undo its damage
      res = 1
      Mid$(FrameData, i, fs - i + 1) = _
        String$(fs - i + 1, Mid$(FrameData, i - 1, 1))
    End If
  End If
  
  i = InStr(FrameData, "LYRICSBEGIN")
  If i Then
    ' how much i hate those 'LYRICS'... grrrrrrrr....
    ' try to restore original data:
    res = 1
    Mid$(FrameData, i, fs - i + 1) = _
      String$(fs - i + 1, Mid$(FrameData, i - 1, 1))
  End If
  
  i = InStr(FrameData, "LAME")
  If i Then
    'fix 'broken' silent frames
    Mid$(FrameData, 5, 32) = String$(32,0)
    res = 1
  End If

  ' Note:
  ' if this is not the last frame, and for example the ID3v1-tag found here
  ' is not complete, it continues in the next frame... so the frameheader is
  ' overwritten...! Thus tidymp3 will not include it in the outputfile.
  ' Nothing to do about that..  :(

  Function = res
End Function


' Returns %TRUE if the passed 128-byte string is a valid ID3v1-tag
Function IsID3v1Tag(ByRef t As String) As Long
  If Mid$(t, 1, 3) = "TAG" Then                      ' starts with "TAG"?
    If IsID3v1TagField(Mid$(t, 4, 30)) Then          ' valid title?
      If IsID3v1TagField(Mid$(t, 34, 30)) Then       ' valid artist?
        If IsID3v1TagField(Mid$(t, 64, 30)) Then     ' valid album?
          If IsID3v1TagField(Mid$(t, 94, 4)) Then    ' valid year?
            If IsID3v1TagField(Mid$(t, 98, 29)) Then ' valid comment?
              Function = %TRUE
            End If
          End If
        End If
      End If
    End If
  End If
End Function


' Returns %TRUE if the specified string is a valid MP3 ID3v1-tag field.
Function IsID3v1TagField(ByRef s As String) As Long
  Local f As Long
  Static i As Long, l As Long, b As Byte, ret As Long
  l = Len(s)
  ret = %TRUE
  For i = 1 To l
    b = Asc(Mid$(s, i, 1))
    If b = 0 Then
      f = %TRUE
    Else
      If (b < 32) Or f Then
        ret = f
        Exit For
      End If
    End If
  Next i
  Function = ret
  ' // Original C-code from CueMaster (mp3scan module):
  ' BOOL f=FALSE;
  ' for(int i=0; i<length; i++){
  '   if((byte)s[i]==0){
  '     f=TRUE;
  '   }else{
  '     if(((byte)s[i]<32)||f==TRUE)return f;
  '   }
  ' }
  ' return TRUE;
End Function
